from dataclasses import dataclass
from .. import types as nt
from typing import Self
from pathlib import Path
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure as mplFigure
from matplotlib.lines import Line2D
from matplotlib.collections import PolyCollection
from matplotlib.container import ErrorbarContainer, BarContainer
from matplotlib.image import AxesImage
from matplotlib.patches import Rectangle
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Line3D
from matplotlib.rcsetup import cycler
from typing import Sequence, TypeVar, Any, Literal
import numpy as np

# TODO: ADD DOCSTRINGS TO PUBLIC METHODS
# TODO: SUPPORT FOR 3D PLOTS
# TODO: SUPPORT FOR FANCY PLOTS (EX: POLAR, IMSHOW, ETC.)

SubplotType = TypeVar("SubplotType", bound="Plot")

color_cycler = cycler(  # type: ignore
    color=[
        "#1f77b4",
        "#aec7e8",
        "#ff7f0e",
        "#ffbb78",
        "#2ca02c",
        "#98df8a",
        "#d62728",
        "#ff9896",
        "#9467bd",
        "#c5b0d5",
        "#8c564b",
        "#c49c94",
        "#e377c2",
        "#f7b6d2",
        "#7f7f7f",
        "#c7c7c7",
        "#bcbd22",
        "#dbdb8d",
        "#17becf",
        "#9edae5",
    ]
)


@dataclass
class PlotSetup:
    """Plot configuration

    PlotSetup is nastro's approach to plot configuration. It is a dataclass including
    most of the features one generally specifies when creating a plot in Matplotlib. This includes aspects such as figure size, labels, scales... but also flags to show or save the figure or to display a legend or a grid. The goal is to simplify the process of creating and customizing plots by concentrating all the configuration in a single object, which can then be used as input for all plotting classes.

    Parameters
    ------------
    figsize : tuple[float, float], optional
        Figure size.
    layout : Literal["tight", "constrained", "none", "compressed"], optional
        Layout used to calculate spacings and sizes of subplots.
    title : str, optional
        Figure title.
    show : bool, optional
        Whether to show the plot or not.
    save : bool, optional
        Whether to save the plot or not.
    dir : str | Path, optional
        Directory to save plot.
    name : str, optional
        Plot filename.
    path: Path, optional
        Full path to save the plot.
    axtitle : str, optional
        Subplot title.
    xlabel : str, optional
        x-axis label.
    ylabel : str, optional
        y-axis label.
    rlabel : str, optional
        Right axis label.
    plabel : str, optional
        Parasite axis label.
    xscale : Literal["linear", "log", "symlog", "logit"], optional
        x-axis scale.
    yscale : Literal["linear", "log", "symlog", "logit"], optional
        y-axis scale.
    rscale : Literal["linear", "log", "symlog", "logit"], optional
        Right axis scale.
    pscale : Literal["linear", "log", "symlog", "logit"], optional
        Parasite axis scale.
    xlim : tuple[float, float], optional
        x-axis limits.
    ylim : tuple[float, float], optional
        y-axis limits.
    rlim : tuple[float, float], optional
        Right axis limits.
    plim : tuple[float, float], optional
        Parasite axis limits.
    xticks : nt.Array, optional
        x-axis ticks.
    yticks : nt.Array, optional
        y-axis ticks.
    rticks : nt.Array, optional
        Right axis ticks.
    pticks : nt.Array, optional
        Parasite axis ticks.
    xticks_idx : nt.Array, optional
        x-axis ticks indices.
    yticks_idx : nt.Array, optional
        y-axis ticks indices.
    rticks_idx : nt.Array, optional
        Right axis ticks indices.
    pticks_idx : nt.Array, optional
        Parasite axis ticks indices.
    colorbar : bool, optional
        Show colorbar.
    cbar_label : str, optional
        Colorbar label.
    cbar_shrink : float, optional
        Colorbar shrink factor.
    legend : bool, optional
        Show legend.
    legend_location : str, optional
        Legend location. Possible values: "best", "upper right", "upper left", "lower left", "lower right", "right", "center left", "center right", "lower center", "upper center", "center".
    grid : bool, optional
        Show grid.
    grid_alpha : float, optional
        Grid transparency.
    """

    # Figure configuration
    figsize: tuple[float, float] = (7, 5)
    layout: Literal["tight", "constrained", "none", "compressed"] = "tight"
    aspect: Literal["auto", "equal"] = "auto"

    title: str | None = None
    show: bool = True
    save: bool = False
    dir: str | Path | None = None
    name: str | None = None
    path: Path | None = None

    # Subplot configuration
    axtitle: str | None = None

    xlabel: str | None = None
    ylabel: str | None = None
    zlabel: str | None = None
    rlabel: str | None = None
    plabel: str | None = None

    xscale: Literal["linear", "log", "symlog", "logit"] = "linear"
    yscale: Literal["linear", "log", "symlog", "logit"] = "linear"
    zscale: Literal["linear", "log", "symlog", "logit"] = "linear"
    rscale: Literal["linear", "log", "symlog", "logit"] = "linear"
    pscale: Literal["linear", "log", "symlog", "logit"] = "linear"

    xlim: tuple[float, float] | None = None
    ylim: tuple[float, float] | None = None
    zlim: tuple[float, float] | None = None
    rlim: tuple[float, float] | None = None
    plim: tuple[float, float] | None = None

    xticks: nt.Array | None = None
    yticks: nt.Array | None = None
    zticks: nt.Array | None = None
    rticks: nt.Array | None = None
    pticks: nt.Array | None = None
    xticks_idx: nt.Array | None = None
    yticks_idx: nt.Array | None = None
    zticks_idx: nt.Array | None = None
    rticks_idx: nt.Array | None = None
    pticks_idx: nt.Array | None = None

    colorbar: bool = True
    cbar_label: str | None = None
    cbar_shrink: float = 1.0

    legend: bool = True
    legend_location: str = "best"
    legend_title: str | None = None
    legend_columns: int = 1

    grid: bool = False
    grid_alpha: float = 0.15

    def __post_init__(self) -> None:

        if self.xticks is not None and self.xticks_idx is None:
            self.xticks_idx = np.arange(len(self.xticks))
        if self.yticks is not None and self.yticks_idx is None:
            self.yticks_idx = np.arange(len(self.yticks))
        if self.zticks is not None and self.zticks_idx is None:
            self.zticks_idx = np.arange(len(self.zticks))
        if self.rticks is not None and self.rticks_idx is None:
            self.rticks_idx = np.arange(len(self.rticks))
        if self.pticks is not None and self.pticks_idx is None:
            self.pticks_idx = np.arange(len(self.pticks))

        if self.save:
            if self.name is None:
                raise ValueError("Missing a filename to save the plot")
            if self.dir is None:
                raise ValueError("Missing a directory to save the plot")

        if self.name is not None:
            if self.dir is None:
                raise ValueError("Missing a directory to save the plot")
            elif isinstance(self.dir, str):
                self.dir = Path(self.dir)

            self.path = self.dir / self.name

        return None

    def copy(self) -> "PlotSetup":
        return PlotSetup(**self.__dict__)


class BaseFigure:

    def __init__(self, setup: PlotSetup, mosaic: str = "a") -> None:

        self.setup = setup

        self.fig = plt.figure(  # type: ignore
            figsize=self.setup.figsize, layout=self.setup.layout
        )
        __subplots = self.fig.subplot_mosaic(mosaic)
        self.subplots = iter(__subplots.values())

        if self.setup.title is not None:
            self.fig.suptitle(self.setup.title)  # type: ignore

        return None


class Plot(BaseFigure):

    def __init__(
        self,
        setup: PlotSetup = PlotSetup(),
        _figure: mplFigure | None = None,
        _axes: Axes | None = None,
    ) -> None:

        if _figure is None and _axes is None:

            self.is_subplot = False
            super().__init__(setup)
            self.ax = next(self.subplots)

        elif _figure is not None and _axes is not None:
            self.is_subplot = True
            self.setup = setup
            self.fig = _figure
            self.ax = _axes
        else:
            raise ValueError("Pass both figure and axes or none of them")

        if self.setup.axtitle is not None:
            self.ax.set_title(self.setup.axtitle)

        if self.setup.xlabel is not None:
            self.ax.set_xlabel(self.setup.xlabel)
        if self.setup.ylabel is not None:
            self.ax.set_ylabel(self.setup.ylabel)

        self.ax.set_xscale(self.setup.xscale)
        self.ax.set_yscale(self.setup.yscale)

        if self.setup.xlim is not None:
            self.ax.set_xlim(self.setup.xlim)
        if self.setup.ylim is not None:
            self.ax.set_ylim(self.setup.ylim)

        if self.setup.grid:
            self.ax.grid(alpha=self.setup.grid_alpha, which="both")

        self.color_cycler = iter(color_cycler)
        self.axes_dict: dict[str, Axes] = {"left": self.ax}
        self.lines: dict[str, tuple[str, Line2D]] = {}
        # self.boundaries: dict[Line2D, PolyCollection] = {}
        # self.error_bars: dict[str, tuple[str, ErrorbarContainer]] = {}
        self.artists: dict[str, tuple[str, Any]] = {}

        return None

    def add_line(
        self,
        x: nt.Array,
        y: nt.Array | None = None,
        fmt: str = "-",
        width: float | None = None,
        markersize: float | None = None,
        color: str | None = None,
        alpha: float = 1.0,
        label: str | None = None,
        axis: str = "left",
    ) -> None:
        """Line

        :param x: Data for the x-axis.
        :param y: Data for the y-axis. If set to None, the data in x will be
            plotted against a range of integers with the same length.
        :param fmt: Line style (no color).
        :param width: Line width.
        :param markersize: Size of the markers.
        :param color: Line color as name or hex code.
        :param alpha: Transparency of the line.
        :param label: Label to identify the line in the legend.
        :param axis: Axis to plot the line.
        """

        if y is None:
            (line,) = self.axes_dict[axis].plot(
                x,
                fmt,
                label=label,
                linewidth=width,
                markersize=markersize,
                alpha=alpha,
            )
        else:
            (line,) = self.axes_dict[axis].plot(
                x,
                y,
                fmt,
                label=label,
                linewidth=width,
                markersize=markersize,
                alpha=alpha,
            )

        name = label if label is not None else f"line{len(self.lines) + 1}"
        self.lines[name] = (axis, line)
        self.artists[name] = (axis, (color, line))

        return None

    def add_scatter(
        self,
        x: nt.Array,
        y: nt.Array | None = None,
        fmt: str = ".",
        width: float | None = None,
        markersize: float | None = 2.0,
        color: str | None = None,
        alpha: float = 1.0,
        label: str | None = None,
        axis: str = "left",
    ) -> None:
        self.add_line(x, y, fmt, width, markersize, color, alpha, label, axis)
        return None

    def add_boundary(
        self,
        limits: nt.Array | nt.Double,
        line: str | Literal["last"] | None = "last",
        follow: bool = False,
        color: str | None = None,
        alpha: float = 0.1,
        label: str | None = None,
        axis: str = "left",
    ) -> None:
        """Solid boundary

        Creates a solid boundary by adding and subtracting a specified limit from
        a reference, which might be a line or just the horizontal axis.

        :param limits: A number or a sequence of numbers representing the limits
            of the boundary.
        :param line: The line to which the boundary will be attached. Can be set
            to "last" to follow the last line added to the plot, to a string
            representing the label of the line that is currently in the plot or
            to None to use the horizontal axis as reference.
        :param follow: Whether the boundary should follow the reference or not.
        :param color: Color of the boundary. If set to None, the boundary will
            have a semi-transparent version of the color of the reference.
        :param alpha: Transparency of the boundary.
        :param label: Label to identify the boundary in the legend.
        :param axis: Axis to plot the boundary.
        """

        if line is None:

            if nt.is_double(limits):
                raise ValueError(
                    "Failed to generate boundary. Missing information about x-axis"
                )

            elif nt.is_array(limits):
                target_axis = axis
                target = None
                limits = np.array(limits, dtype=np.float64)
                x = np.arange(len(limits))
                reference = np.zeros_like(x)

            else:
                raise ValueError(
                    "Failed to generate boundary."
                    " Limits must be a number or a sequence"
                )

        else:

            if line == "last":
                target_axis, target = list(self.lines.values())[-1]
            else:
                if line not in self.lines.keys():
                    raise ValueError(
                        "Failed to generate boundary. "
                        f"Line {line} not found in the plot"
                    )
                target_axis, target = self.lines[line]

            x = target.get_xdata()
            reference = target.get_ydata() if follow else np.zeros_like(x)

            if nt.is_array(limits):
                limits = np.array(limits)
                if limits.shape != np.array(x).shape:
                    raise ValueError(
                        "Failed to generate boundary."
                        " Limits and line have incompatible shapes."
                    )
            else:
                limits = np.array(limits)

        boundary = self.axes_dict[target_axis].fill_between(
            x,
            reference + limits,
            reference - limits,
            alpha=alpha,
            label=label,
        )

        out = (target_axis, (target, color, boundary))

        if label is not None:
            self.artists[label] = out
        else:
            count = len(self.artists) + 1
            self.artists[f"bound-{count}"] = out

        return None

    def add_vertical_boundary(
        self,
        low: nt.Double,
        high: nt.Double,
        color: str | None = None,
        alpha: float = 0.1,
        label: str | None = None,
        axis: str = "left",
    ) -> None:
        """Vertical boundary

        Generates a colored region between two vertical lines spanning the entire
        height of the plot.

        :param low: Lower limit of the boundary.
        :param high: Upper limit of the boundary.
        :param color: Color of the boundary.
        :param alpha: Transparency of the boundary.
        :param label: Label to identify the boundary in the legend.
        :param axis: Axis to plot the boundary.
        """

        boundary = self.axes_dict[axis].add_artist(
            Rectangle(
                (low, -1e20),  # type: ignore
                high - low,  # type: ignore
                2e20,
                alpha=alpha,
                color=color,
                label=label,
            )
        )

        out = (axis, (None, color, boundary))
        name = label if label is not None else f"vbound-{len(self.artists) + 1}"
        self.artists[name] = out

        return None

    def add_errorbar(
        self,
        x: nt.Array,
        y: nt.Array,
        error: nt.Array,
        fmt: str = "-",
        color: str | None = None,
        label: str | None = None,
        axis: str = "left",
    ) -> None:

        errorbar = self.axes_dict[axis].errorbar(
            x, y, yerr=error, fmt=fmt, label=label, color=color
        )

        if label is not None:
            # self.error_bars[label] = (axis, errorbar)
            self.artists[label] = (axis, errorbar)
        else:
            __number = len(self.artists) + 1
            # self.error_bars[f"errorbar{__number}"] = (axis, errorbar)
            self.artists[f"artist{__number}"] = (axis, errorbar)

        return None

    def add_step(
        self,
        x: nt.Array,
        y: nt.Array | None = None,
        where: Literal["pre", "post", "mid"] = "pre",
        fmt: str = "-",
        color: str | None = None,
        label: str | None = None,
        axis: str = "left",
    ) -> None:

        if y is None:
            step = self.axes_dict[axis].step(
                x, fmt, where=where, label=label, color=color
            )
        else:
            step = self.axes_dict[axis].step(
                x, y, fmt, where=where, label=label, color=color
            )

        if label is not None:
            self.artists[label] = (axis, step)
        else:
            __number = len(self.artists) + 1
            self.artists[f"artist{__number}"] = (axis, step)

        return None

    def add_barplot(
        self,
        x: nt.Array,
        height: nt.Array,
        width: float = 0.8,
        ticks: nt.Array | None = None,
        axis: str = "left",
    ) -> None:

        bar = self.axes_dict[axis].bar(x, height, width=width, tick_label=ticks)
        count = len(self.artists) + 1
        self.artists[f"bar-{count}"] = (axis, bar)

        return None

    def add_horizontal_barplot(
        self,
        x: nt.Array,
        width: nt.Array,
        height: float = 0.8,
        ticks: nt.Array | None = None,
        axis: str = "left",
    ) -> None:

        bar = self.axes_dict[axis].barh(x, width, height=height, tick_label=ticks)
        count = len(self.artists) + 1
        self.artists[f"hbar-{count}"] = (axis, bar)

        return None

    def add_histogram(
        self,
        data: nt.Array,
        bins: int = 10,
        normalize: bool = False,
        cumulative: bool = False,
        hist_type: Literal["bar", "step", "stepfilled"] = "bar",
        align: Literal["left", "mid", "right"] = "mid",
        label: str | None = None,
        alpha: float = 0.8,
        axis: str = "left",
    ) -> None:
        """Histogram

        :param data: A sequence of numbers.
        :param bins: Number of intervals in which the data is divided.
        :param normalize: If False, the vertical axis represents the number of
            elements in each bin. If True, the counts are normalized by the total
            number of elements, so that the vertical axis represents the
            probability of a number belonging to each interval.
        :param cumulative: If True, each bin contains the number of elements that
            are smaller or equal to its upper limit.
        :param hist_type: Bar plots a sequence of rectangles, step plots the line
            joining the top of the rectangles, stepfilled plots the same line as
            step but fills the area below it.
        :param align: Whether bars will be centered on the left edge, right edge
            or in the middle of the intervals.
        :param label: To be shown in the legend to identify the histogram.
        :param alpha: Transparency of the bars.
        :param axis: Axis to plot the histogram.
        """

        _, _, hist = self.axes_dict[axis].hist(
            data,
            bins=bins,
            density=normalize,
            cumulative=cumulative,
            histtype=hist_type,
            align=align,
            label=label,
            alpha=alpha,
        )

        count = len(self.artists) + 1
        self.artists[f"hist-{count}"] = (axis, hist)

        return None

    def add_image(self, data: nt.Vector, cmap: str = "GnBu") -> None:

        if len(data.shape) != 2:
            raise ValueError("Data must be a 2D array")

        if data.shape[0] != data.shape[1]:
            raise ValueError("Data must be a square matrix")

        image = self.axes_dict["left"].imshow(data, cmap=cmap)
        __number = len(self.artists) + 1
        self.artists[f"artist{__number}"] = ("left", image)

        return None

    def next_color(self) -> str:
        try:
            return next(self.color_cycler)["color"]
        except StopIteration:
            self.color_cycler = iter(color_cycler)
            return next(self.color_cycler)["color"]

    def postprocess(self) -> None:

        # TODO: REFACTOR COLORING

        # Ticks
        if self.setup.xticks is not None and self.setup.xticks_idx is not None:
            self.ax.set_xticks(self.setup.xticks_idx, self.setup.xticks)
        if self.setup.yticks is not None and self.setup.yticks_idx is not None:
            self.ax.set_yticks(self.setup.yticks_idx, self.setup.yticks)
        if "right" in self.axes_dict.keys():
            if self.setup.rticks is not None and self.setup.rticks_idx is not None:
                self.axes_dict["right"].set_yticks(
                    self.setup.rticks_idx, self.setup.rticks
                )
        if "parasite" in self.axes_dict.keys():
            if self.setup.pticks is not None and self.setup.pticks_idx is not None:
                self.axes_dict["parasite"].set_yticks(
                    self.setup.pticks_idx, self.setup.pticks
                )

        # Colors
        for key, (_, artist) in self.artists.items():

            # Line
            if isinstance(artist, tuple) and isinstance(artist[-1], Line2D):
                color, line = artist
                if color is None:
                    color = self.next_color()
                line.set_color(color)

            # if isinstance(artist, Line2D):
            #     artist.set_color(self.next_color())

            elif isinstance(artist, ErrorbarContainer):
                color = self.next_color()
                artist[0].set_color(color)
                for tick in artist[2]:
                    tick.set_color(color)

            elif isinstance(artist, BarContainer):

                if "hist" in key:
                    color = self.next_color()
                    for bar in artist:
                        bar.set_color(color)
                else:
                    for bar in artist:
                        bar.set_color(self.next_color())

            # Vertical boundary
            elif isinstance(artist, tuple) and isinstance(artist[2], Rectangle):
                target, color, boundary = artist
                if color is None:
                    color = self.next_color() if target is None else target.get_color()
                boundary.set_color(color)

            elif isinstance(artist, list):
                if isinstance(artist[0], Line2D):
                    color = self.next_color()
                    for line in artist:
                        line.set_color(color)
                else:
                    try:
                        color = self.next_color()
                        for item in artist:
                            item.set_color(color)
                    except:
                        raise ValueError(f"Failed to color artist: {type(artist[0])}")

            elif isinstance(artist, AxesImage):

                if not self.setup.colorbar:
                    continue

                self.fig.colorbar(
                    artist,
                    ax=self.ax,
                    label=self.setup.cbar_label,
                    shrink=self.setup.cbar_shrink,
                )

            # Boundary
            elif isinstance(artist, tuple) and isinstance(artist[-1], PolyCollection):
                target, color, boundary = artist
                if color is None:
                    color = self.next_color() if target is None else target.get_color()
                boundary.set_color(color)
            else:
                print(f"Unknown artist type: {type(artist)}")

        # Legend
        __handles = []
        for axis in self.axes_dict.values():
            for handle in axis.get_legend_handles_labels()[0]:
                __handles.append(handle)
        if self.setup.legend and len(__handles) > 0:
            last_axis = list(self.axes_dict.values())[-1]
            last_axis.legend(
                loc=self.setup.legend_location,
                handles=__handles,
                title=self.setup.legend_title,
                ncols=self.setup.legend_columns,
            )

        # Aspect ratio
        for axis in self.axes_dict.values():
            axis.set_aspect(self.setup.aspect)

        return None

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:

        self.postprocess()

        if "right" in self.axes_dict.keys():
            if len(self.axes_dict["right"].get_lines()) > 1:
                raise ValueError("Attempted to plot multiple lines on the right axis")
            line = self.axes_dict["right"].get_lines()[-1]
            self.axes_dict["right"].yaxis.label.set_color(line.get_color())

        if "parasite" in self.axes_dict.keys():
            if len(self.axes_dict["parasite"].get_lines()) > 1:
                raise ValueError(
                    "Attempted to plot multiple lines on the parasite axis"
                )
            line = self.axes_dict["parasite"].get_lines()[-1]
            self.axes_dict["parasite"].yaxis.label.set_color(line.get_color())

        if self.is_subplot:
            return None

        if self.setup.save:
            assert isinstance(self.setup.path, Path)
            self.setup.path.parent.mkdir(parents=True, exist_ok=True)
            self.fig.savefig(self.setup.path)

        if self.setup.show:
            plt.show(block=True)

        plt.close("all")

        return None


class SingleAxis(Plot):
    """A 2D plot with a single vertical axis"""

    pass


class DoubleAxis(Plot):
    """A 2D plot with left and right vertical axes"""

    def __init__(
        self,
        setup: PlotSetup = PlotSetup(),
        _figure: mplFigure | None = None,
        _axes: Axes | None = None,
    ) -> None:

        super().__init__(setup, _figure, _axes)

        # Add right axis
        self.right = self.ax.twinx()
        assert isinstance(self.right, Axes)
        self.axes_dict["right"] = self.right

        if self.setup.rlabel is not None:
            self.right.set_ylabel(self.setup.rlabel)
        self.right.set_yscale(self.setup.rscale)
        if self.setup.rlim is not None:
            self.right.set_ylim(self.setup.rlim)

        return None


class ParasiteAxis(DoubleAxis):
    """A 2D plot with left, right and parasite vertical axes"""

    def __init__(
        self,
        setup: PlotSetup = PlotSetup(),
        _figure: mplFigure | None = None,
        _axes: Axes | None = None,
    ) -> None:

        super().__init__(setup, _figure, _axes)

        # Add parasite axis
        self.parax = self.ax.twinx()
        assert isinstance(self.parax, Axes)
        self.parax.spines.right.set_position(("axes", 1.13))
        self.axes_dict["parasite"] = self.parax

        if self.setup.plabel is not None:
            self.parax.set_ylabel(self.setup.plabel)
        self.parax.set_yscale(self.setup.pscale)
        if self.setup.plim is not None:
            self.parax.set_ylim(self.setup.plim)

        return None


class Mosaic(BaseFigure):
    """A figure with multiple subplots arranged in a grid"""

    def __init__(self, mosaic: str, setup: PlotSetup = PlotSetup()) -> None:
        super().__init__(setup, mosaic)

    def add_subplot(
        self,
        setup: PlotSetup = PlotSetup(),
        generator: type[SubplotType] = SingleAxis,
    ) -> SubplotType:
        return generator(setup, self.fig, next(self.subplots))

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:

        if exc_type is not None:
            plt.close(self.fig)
            raise exc_type(exc_value).with_traceback(exc_traceback)

        if self.setup.save:
            assert isinstance(self.setup.path, Path)
            self.setup.path.parent.mkdir(parents=True, exist_ok=True)
            self.fig.savefig(self.setup.path)

        if self.setup.show:
            plt.show(block=True)

        plt.close("all")
        return None


class Base3D:

    def __init__(self, setup: PlotSetup = PlotSetup()) -> None:

        self.setup = setup

        # Create figure and axis
        self.fig = plt.figure(figsize=self.setup.figsize)
        self.ax = self.fig.add_subplot(
            projection="3d", proj_type="ortho", box_aspect=(1, 1, 1)
        )

        # Set title and labels
        if self.setup.title is not None:
            self.fig.suptitle(self.setup.title)

        if self.setup.xlabel is not None:
            self.ax.set_xlabel(self.setup.xlabel)
        if self.setup.xscale is not None:
            self.ax.set_xscale(self.setup.xscale)
        if self.setup.xlim is not None:
            self.ax.set_xlim(self.setup.xlim)

        if self.setup.ylabel is not None:
            self.ax.set_ylabel(self.setup.ylabel)
        if self.setup.yscale is not None:
            self.ax.set_yscale(self.setup.yscale)
        if self.setup.ylim is not None:
            self.ax.set_ylim(self.setup.ylim)

        if self.setup.zlabel is not None:
            self.ax.set_zlabel(self.setup.zlabel)  # type: ignore
        if self.setup.zscale is not None:
            self.ax.set_zscale(self.setup.zscale)  # type: ignore
        if self.setup.zlim is not None:
            self.ax.set_zlim(self.setup.zlim)  # type: ignore

        self.color_cycler = iter(color_cycler)
        self.axes_dict: dict[str, Axes] = {"left": self.ax}
        self.artists: dict[str, tuple[str, Any]] = {}

        return None

    def next_color(self) -> str:
        try:
            return next(self.color_cycler)["color"]
        except StopIteration:
            self.color_cycler = iter(color_cycler)
            return next(self.color_cycler)["color"]

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.postprocess()

    def add_line(
        self,
        x: nt.Array,
        y: nt.Array,
        z: nt.Array,
        fmt: str = "-",
        width: float | None = None,
        markersize: float | None = None,
        color: str | None = None,
        alpha: float = 1.0,
        label: str | None = None,
        axis: str = "left",
    ) -> None:
        """Line

        :param x: Data for the x-axis.
        :param y: Data for the y-axis. If set to None, the data in x will be
            plotted against a range of integers with the same length.
        :param fmt: Line style (no color).
        :param width: Line width.
        :param markersize: Size of the markers.
        :param color: Line color as name or hex code.
        :param alpha: Transparency of the line.
        :param label: Label to identify the line in the legend.
        """

        (line,) = self.axes_dict[axis].plot(
            x,
            y,
            z,
            fmt,
            label=label,
            linewidth=width,
            markersize=markersize,
            alpha=alpha,
        )

        name = label if label is not None else f"line{len(self.artists) + 1}"
        self.artists[name] = (axis, (color, line))

        return None

    def postprocess(self) -> None:

        original_limits = np.array(
            [
                self.ax.get_xlim(),
                self.ax.get_ylim(),
                self.ax.get_zlim(),  # type: ignore
            ]
        ).T

        homogeneous_limits = (np.min(original_limits[0]), np.max(original_limits[1]))

        self.ax.set_xlim(homogeneous_limits)
        self.ax.set_ylim(homogeneous_limits)
        self.ax.set_zlim(homogeneous_limits)  # type: ignore

        # Ticks
        if self.setup.xticks is not None and self.setup.xticks_idx is not None:
            self.ax.set_xticks(self.setup.xticks_idx, self.setup.xticks)
        if self.setup.yticks is not None and self.setup.yticks_idx is not None:
            self.ax.set_yticks(self.setup.yticks_idx, self.setup.yticks)

        # Colors
        for key, (_, artist) in self.artists.items():

            # Line
            if isinstance(artist, tuple) and isinstance(artist[-1], Line3D):
                color, line = artist
                if color is None:
                    color = self.next_color()
                line.set_color(color)
            else:
                print(f"Unknown artist type: {type(artist)}")

        # Legend
        __handles = []
        for axis in self.axes_dict.values():
            for handle in axis.get_legend_handles_labels()[0]:
                __handles.append(handle)
        if self.setup.legend and len(__handles) > 0:
            last_axis = list(self.axes_dict.values())[-1]
            last_axis.legend(
                loc=self.setup.legend_location,
                handles=__handles,
                title=self.setup.legend_title,
                ncols=self.setup.legend_columns,
            )

        if self.setup.save:
            assert isinstance(self.setup.path, Path)
            self.setup.path.parent.mkdir(parents=True, exist_ok=True)
            self.fig.savefig(self.setup.path)

        if self.setup.show:
            plt.show(block=True)

        plt.close("all")

        return None


class Plot3D(Base3D):
    pass


# class Base3D:
#     def __init__(self, setup: PlotSetup = PlotSetup()) -> None:
#         self.setup = setup

#         # Create figure and axis
#         self.fig = plt.figure(figsize=self.setup.figsize)
#         self.ax = self.fig.add_subplot(
#             projection="3d", proj_type="ortho", box_aspect=(1, 1, 1)
#         )

#         # Set title and labels
#         if self.setup.title is not None:
#             self.fig.suptitle(self.setup.title, fontsize="x-large")

#         if self.setup.xlabel is not None:
#             self.ax.set_xlabel(self.setup.xlabel)
#         if self.setup.xscale is not None:
#             self.ax.set_xscale(self.setup.xscale)
#         if self.setup.xlim is not None:
#             self.ax.set_xlim(self.setup.xlim)

#         if self.setup.ylabel is not None:
#             self.ax.set_ylabel(self.setup.ylabel)
#         if self.setup.yscale is not None:
#             self.ax.set_yscale(self.setup.yscale)
#         if self.setup.ylim is not None:
#             self.ax.set_ylim(self.setup.ylim)

#         if self.setup.zlabel is not None:
#             self.ax.set_zlabel(self.setup.zlabel)  # type: ignore
#         if self.setup.zscale is not None:
#             self.ax.set_zscale(self.setup.zscale)  # type: ignore
#         if self.setup.zlim is not None:
#             self.ax.set_zlim(self.setup.zlim)  # type: ignore

#         self.path = ""

#         return None

#     def __enter__(self):
#         return self

#     def __exit__(self, exc_type, exc_value, exc_traceback):
#         self.postprocess()

#     def add_line(
#         self, x: nt.Array, y: Array, z: Array, fmt: str = "-", label: str | None = None
#     ) -> None:
#         self.ax.plot(x, y, z, fmt, label=label)

#         return None

#     def postprocess(self) -> str:
#         original_limits = np.array(
#             [
#                 self.ax.get_xlim(),
#                 self.ax.get_ylim(),
#                 self.ax.get_zlim(),  # type: ignore
#             ]
#         ).T

#         homogeneous_limits = (np.min(original_limits[0]), np.max(original_limits[1]))

#         self.ax.set_xlim(homogeneous_limits)
#         self.ax.set_ylim(homogeneous_limits)
#         self.ax.set_zlim(homogeneous_limits)  # type: ignore

#         labels = self.ax.get_legend_handles_labels()[1]
#         if self.setup.legend and len(labels) > 0:
#             self.ax.legend()

#         if self.setup.save:
#             Path(self.setup.dir).mkdir(parents=True, exist_ok=True)
#             self.path = f"{self.setup.dir}/{self.setup.name}"
#             self.fig.savefig(self.path)

#         if self.setup.show:
#             plt.show(block=True)
#             plt.close(self.fig)

#         return self.path


# class Plot3D(Base3D):
#     def plot(
#         self, x: nt.Array, y: Array, z: Array, fmt: str = "-", label: str | None = None
#     ) -> str:
#         self.add_line(x, y, z, fmt=fmt, label=label)
#         self.postprocess()
#         return self.path



from .base import CartesianPropagator, SimpleCartesianPropagator


class SimpleRK4(SimpleCartesianPropagator):

    def _step_impl(self) -> bool:

        new_time = self.now + self.h

        k1 = self.model(self.now, self.s)
        k2 = self.model(self.now + 0.5 * self.h, self.s + k1.times_dt(0.5 * self.h))
        k3 = self.model(self.now + 0.5 * self.h, self.s + k2.times_dt(0.5 * self.h))
        k4 = self.model(self.now + self.h, self.s + k3.times_dt(self.h))
        sum_k = (k1 + k2 * 2.0 + k3 * 2.0 + k4) / 6.0
        self.s = self.s + sum_k.times_dt(self.h)
        self.now = new_time

        return True


class RungeKuttaPropagator(CartesianPropagator):

    def __init__(self, fun, t0, s0, tend, dt) -> None:

        super().__init__(fun, t0, s0, tend, dt)

        return None

    def _step_impl(self) -> bool:

        # TODO: Handle units properly (???? Apr 2024)

        new_time_int, new_time_fr = self._update_epoch()
        h_int, h_fr = self._get_steps()

        k1 = self.model(self.now_int, self.s, fr=self.now_fr)
        k2 = self.model(
            self.now_int + 0.5 * h_int,
            self.s + k1.times_dt(0.5 * h_int) + k1.times_dt(0.5 * h_fr),
            fr=self.now_fr + 0.5 * h_fr,
        )
        k3 = self.model(
            self.now_int + 0.5 * h_int,
            self.s + k2.times_dt(0.5 * h_int) + k2.times_dt(0.5 * h_fr),
            fr=self.now_fr + 0.5 * h_fr,
        )
        k4 = self.model(
            self.now_int + self.h_int,
            self.s + k3.times_dt(h_int) + k3.times_dt(h_fr),
            fr=self.now_fr + self.h_fr,
        )
        sum_k = (k1 + k2 * 2.0 + k3 * 2.0 + k4) * (1.0 / 6.0)

        new_s = self.s + sum_k.times_dt(h_int) + sum_k.times_dt(h_fr)

        self.now_int = new_time_int
        self.now_fr = new_time_fr
        self.s = new_s

        return True



from ..types import Double, Vector
from typing import Any, TypeVar
from ..catalog import Earth
import numpy as np

# from .state import CartesianPosition, SphericalPosition
from ..constants import miliarcsec

# from typing import Generic

Frame = TypeVar("Frame", bound="ReferenceFrame")


class ReferenceFrame[T: (Double, Vector)]:
    """Base class for reference frames"""

    variables = {
        "rho": "q1",
        "alpha": "q2",
        "delta": "q3",
        "drho": "q4",
        "dalpha": "q5",
        "ddelta": "q6",
    }
    angles: dict[str, tuple[float, float]] = {
        "q2": (0.0, 2 * np.pi),
        "q3": (-0.5 * np.pi, 0.5 * np.pi),
    }

    @classmethod
    def to_cartesian(cls, q1: T, q2: T, q3: T) -> tuple[T, T, T]:
        _q1: Any = q1 * np.cos(q2) * np.cos(q3)
        _q2: Any = q1 * np.sin(q2) * np.cos(q3)
        _q3: Any = q1 * np.sin(q3)
        return _q1, _q2, _q3

    @classmethod
    def to_spherical(cls, q1: T, q2: T, q3: T) -> tuple[T, T, T]:
        _q1: Any = np.sqrt(q1 * q1 + q2 * q2 + q3 * q3)
        _q2: Any = np.arctan2(q2, q1)
        _q3: Any = np.arcsin(q3 / _q1)
        return _q1, _q2, _q3

    @classmethod
    def transform(cls, q1: T, q2: T, q3: T, frame: type[Frame]) -> tuple[T, T, T]:
        """Transform cartesian vector to a different reference frame"""

        try:
            return getattr(cls, f"to_{frame.__name__.lower()}")(q1, q2, q3)
        except AttributeError:
            raise NotImplementedError(
                f"Conversion from {cls.__name__} to {frame.__name__} not implemented"
            )

    @classmethod
    def to_icrf(cls, q1: T, q2: T, q3: T) -> tuple[T, T, T]:
        raise NotImplementedError(
            "Attempted conversion to ICRF from generic reference frame"
        )

    @classmethod
    def to_j2000(cls, q1: T, q2: T, q3: T) -> tuple[T, T, T]:
        raise NotImplementedError(
            "Attempted conversion to J2000 from generic reference frame"
        )


class ICRF[T: (Double, Vector)](ReferenceFrame):
    """International Celestial Reference Frame"""

    dalpha0 = -14.6 * miliarcsec
    xi0 = -16.6170 * miliarcsec
    eta0 = -6.8192 * miliarcsec

    B = np.array(
        [
            [
                1.0 - 0.5 * (dalpha0 * dalpha0 + xi0 * xi0),
                dalpha0,
                -xi0,
            ],
            [
                -dalpha0 - eta0 * xi0,
                1.0 - 0.5 * (dalpha0 * dalpha0 + eta0 * eta0),
                -eta0,
            ],
            [
                xi0 - eta0 * dalpha0,
                eta0 + xi0 * dalpha0,
                1.0 - 0.5 * (xi0 * xi0 + eta0 * eta0),
            ],
        ]
    )

    @classmethod
    def to_icrf(cls, q1: T, q2: T, q3: T) -> tuple[T, T, T]:
        return q1, q2, q3

    @classmethod
    def to_j2000(cls, q1: T, q2: T, q3: T) -> tuple[T, T, T]:
        """Convert from ICRF to J2000 reference frame"""
        q1, q2, q3 = np.dot(cls.B, np.array([q1, q2, q3]))
        return q1, q2, q3


class J2000[T: (Double, Vector)](ReferenceFrame):
    """Mean Equator and Equinox of J2000"""

    @classmethod
    def to_icrf(cls, q1: T, q2: T, q3: T) -> tuple[T, T, T]:
        """Convert from J2000 to ICRF reference frame"""
        q1, q2, q3 = np.dot(ICRF.B.T, np.array([q1, q2, q3]))
        return q1, q2, q3

    @classmethod
    def to_j2000(cls, q1: T, q2: T, q3: T) -> tuple[T, T, T]:
        return q1, q2, q3


class ITRF[T: (Double, Vector)](ReferenceFrame):
    """International Terrestrial Reference Frame"""

    variables = {
        "h": "q1",
        "lon": "q2",
        "lat": "q3",
        "dh": "q4",
        "dlon": "q5",
        "dlat": "q6",
    }

    @classmethod
    def to_cartesian(cls, q1: T, q2: T, q3: T) -> tuple[T, T, T]:
        """Conversion from geodetic spherical coordinates to cartesian"""

        # Earth flattening and squared eccentricity
        f = (Earth.Re - Earth.Rp) / Earth.Re
        e2 = 2.0 * f - f * f

        # Ecliptic and out-of-plane components
        sinq2 = np.sin(q2)
        C = Earth.Re / np.sqrt(1.0 - e2 * sinq2 * sinq2)
        S = C * (1.0 - e2)

        # Cartesian coordinates
        common = (C + q1) * np.cos(q2)
        cartesian_q1: Any = common * np.cos(q3)
        cartesian_q2: Any = common * np.sin(q3)
        cartesian_q3: Any = (S + q1) * sinq2

        return cartesian_q1, cartesian_q2, cartesian_q3

    @classmethod
    def to_spherical(cls, q1: T, q2: T, q3: T) -> tuple[T, T, T]:
        """Conversion from cartesian to geodetic spherical coordinates"""

        # Longitude and recurrent parameters
        sph_q3: Any = np.arctan2(q2, q1)
        f = (Earth.Re - Earth.Rp) / Earth.Re
        e2 = 2 * f - f * f
        p = np.sqrt(q1 * q1 + q2 * q2)

        # Iterate to find height and latitude
        sph_q2: Any = 0.0 * sph_q3
        dsph_q2 = sph_q2 + 1.0
        sph_q1 = sph_q2
        tol = 1e-15

        while np.any(np.abs(dsph_q2) > tol):
            shp_q2_old = sph_q2
            N = Earth.Re / np.sqrt(1.0 - e2 * np.sin(sph_q2) * np.sin(sph_q2))
            sph_q1 = p / np.cos(sph_q2) - N
            sph_q2 = np.arctan2(q3 * (N + sph_q1), p * (N + sph_q1 - e2 * N))
            dsph_q2 = sph_q2 - shp_q2_old

        return sph_q1, sph_q2, sph_q3


# class ReferenceFrame[T: (Double, Vector)]:
#     """Base class for reference frames"""

#     variables = {"r": "q1", "theta": "q2", "phi": "q3"}

#     @classmethod
#     def convert(cls, frame: type[Frame], *qs: T) -> tuple[T, ...]:
#         """Convert position vector to a different reference frame

#         :param r: Position vector in current frame
#         :param frame: Reference frame to convert to
#         """
#         if frame == ICRF:
#             return cls.__perform_conversion(cls.to_icrf, *qs)
#         elif frame == J2000:
#             return cls.__perform_conversion(cls.to_j2000, *qs)

#         raise NotImplementedError(
#             f"Conversion from {cls.__name__} to {frame.__name__} not implemented"
#         )

#     @classmethod
#     def __perform_conversion(cls, func, *qs: T) -> tuple[T, ...]:

#         if len(qs) == 3:
#             return func(*qs)
#         elif len(qs) == 6:
#             raise NotImplementedError("Conversion of velocity not implemented")
#         else:
#             raise ValueError("Invalid number of coordinates")

#     @classmethod
#     def to_icrf(cls, *qs: T) -> tuple[T, ...]:
#         raise NotImplementedError

#     @classmethod
#     def to_j2000(cls, *qs: T) -> tuple[T, ...]:
#         raise NotImplementedError

#     @classmethod
#     def to_gcrf(cls, *qs: T) -> tuple[T, ...]:
#         raise NotImplementedError

#     @classmethod
#     def to_spherical(cls, x: T, y: T, z: T) -> tuple[T, T, T]:
#         raise NotImplementedError

#     @classmethod
#     def to_cartesian(cls, r: T, theta: T, phi: T) -> tuple[T, T, T]:
#         raise NotImplementedError


# class ICRF[T: (Double, Vector)](ReferenceFrame):
#     """International Celestial Reference Frame

#     Inertial reference frame with origin at the barycenter of the solar system
#     and orientation defined from extragalactic sources. The ICRF is aligned with
#     the J2000 equatorial frame (Equator as XY plane and vernal equinox as X axis)
#     to within a few milliarcseconds.

#     Attributes
#     ----------
#     B : np.ndarray
#         Frame-bias matrix, used to convert from ICRF to J2000.

#     Sources
#     -------
#     https://arxiv.org/abs/astro-ph/0602086
#     """

#     variables = {"r": "q1", "alpha": "q2", "delta": "q3"}

#     dalpha0 = -14.6 * miliarcsec
#     xi0 = -16.6170 * miliarcsec
#     eta0 = -6.8192 * miliarcsec

#     B = np.array(
#         [
#             [
#                 1.0 - 0.5 * (dalpha0 * dalpha0 + xi0 * xi0),
#                 dalpha0,
#                 -xi0,
#             ],
#             [
#                 -dalpha0 - eta0 * xi0,
#                 1.0 - 0.5 * (dalpha0 * dalpha0 + eta0 * eta0),
#                 -eta0,
#             ],
#             [
#                 xi0 - eta0 * dalpha0,
#                 eta0 + xi0 * dalpha0,
#                 1.0 - 0.5 * (xi0 * xi0 + eta0 * eta0),
#             ],
#         ]
#     )

#     @classmethod
#     def to_icrf(cls, *qs: T) -> tuple[T, ...]:
#         return qs

#     @classmethod
#     def to_j2000(cls, *qs: T) -> tuple[T, ...]:
#         """Convert from ICRF to J2000 reference frame"""
#         out = np.dot(cls.B, np.array(qs))
#         return out[0], out[1], out[2]


# class J2000[T: (Double, Vector)](ReferenceFrame):
#     """Mean Equator and Equinox of J2000

#     Former standard reference frame for celestial coordinates, related to the
#     ICRF by the frame-bias matrix (B). The XY plane is the mean equator and the
#     X axis is the mean equinox of J2000.

#     Sources
#     -------
#     https://arxiv.org/abs/astro-ph/0602086
#     """

#     @classmethod
#     def to_j2000(cls, *qs: T) -> tuple[T, ...]:
#         return qs

#     @classmethod
#     def to_icrf(cls, *qs: T) -> tuple[T, ...]:
#         """Convert from J2000 to ICRF reference frame"""
#         out = np.dot(ICRF.B.T, np.array(qs))
#         return out[0], out[1], out[2]


# class GCRF[T: (Double, Vector)](ReferenceFrame):
#     """Geocentric Celestial Reference Frame

#     Geocentric version of ICRF
#     """

#     variables = {"r": "q1", "alpha": "q2", "delta": "q3"}


# class ITRF[T: (Double, Vector)](ReferenceFrame):
#     """International Terrestrial Reference Frame"""

#     variables = {"h": "q1", "lat": "q2", "lon": "q3"}

#     @classmethod
#     def to_cartesian(cls, r: T, theta: T, phi: T) -> tuple[T, T, T]:

#         height, lat, lon = r, theta, phi

#         # Compute flattening and eccentricity
#         f = (Earth.Re - Earth.Rp) / Earth.Re
#         e2 = 2 * f - f * f

#         # Compute ecliptic plane and out-of-plane components
#         sinlat = np.sin(lat)
#         C = Earth.Re / np.sqrt(1.0 - e2 * sinlat * sinlat)
#         S = C * (1.0 - e2)

#         # Compute Cartesian coordinates
#         rd = (C + height) * np.cos(lat)
#         x: Any = rd * np.cos(lon)
#         y: Any = rd * np.sin(lon)
#         z: Any = (S + height) * sinlat

#         return x, y, z

#     @classmethod
#     def to_spherical(cls, x: T, y: T, z: T) -> tuple[T, T, T]:

#         # Compute longitude and recurrent parameters
#         lon = np.arctan2(y, x)
#         f = (Earth.Re - Earth.Rp) / Earth.Re
#         e2 = 2 * f - f * f
#         p = np.sqrt(x * x + y * y)

#         # Iterate to find height and latitude
#         lat = 0.0 * lon
#         dlat = lat + 1.0
#         height = lat
#         tol = 1e-12

#         while np.any(np.abs(dlat) > tol):
#             lat0 = lat
#             N = Earth.Re / np.sqrt(1.0 - e2 * np.sin(lat) * np.sin(lat))
#             height = p / np.cos(lat) - N
#             lat = np.arctan2(z * (N + height), p * (N + height - e2 * N))
#             dlat = lat - lat0

#         height: Any = height
#         lat: Any = lat
#         lon: Any = lon

#         return height, lat, lon



from .core import Vector, Double
from typing import Any
import numpy as np


def time_to_mean_anomaly[T](time: T, a: Double, mu: Double, M0: Double = 0.0) -> T:
    """Mean anomaly from time past periapsis.

    :param time: Time since periapsis [s]
    :param a: Semi-major axis of the keplerian orbit [m]
    :param mu: Standard gravitational parameter of central body [m^3/s^2]
    :param M0: Initial mean anomaly [rad]
    """
    return M0 + time * np.sqrt(mu / (a * a * a))


def true_to_eccentric_anomaly[T: (Double, Vector)](ta: T, e: T) -> T:
    """Eccentric anomaly from true anomaly and eccentricity

    :param theta: True anomaly [rad]
    :param e: Eccentricity of the keplerian orbit
    """
    out: Any = 2.0 * np.arctan2(
        np.sqrt(1.0 - e) * np.sin(0.5 * ta), np.sqrt(1.0 + e) * np.cos(0.5 * ta)
    )
    return out


def eccentric_to_true_anomaly[T: (Double, Vector)](E: T, e: T) -> T:
    """True anomaly from eccentric anomaly and eccentricity

    :param E: Eccentric anomaly [rad]
    :param e: Eccentricity of the keplerian orbit
    """
    out: Any = 2.0 * np.arctan2(
        np.sqrt(1.0 + e) * np.sin(0.5 * E), np.sqrt(1.0 - e) * np.cos(0.5 * E)
    )
    return out


def eccentric_to_mean_anomaly[T: (Double, Vector)](E: T, e: T) -> T:
    """Mean anomaly from eccentric anomaly and eccentricity

    :param E: Eccentric anomaly [rad]
    :param e: Eccentricity of the keplerian orbit
    """
    out: Any = E - e * np.sin(E)
    return out


# def cartesian2spherical(t: Vector, s: CartesianState) -> SphericalGeocentric:
#     """Convert cartesian state vectors to spherical coordinates.

#     :param t: Epoch in which the cartesian state is knonw or time series [JD].
#     :param s: Cartesian state [m, m/s].
#     :return: Spherical coordinates [m, deg].
#     """

#     def _constrain_theta(theta: float) -> float:
#         """Constain angle to [0, 360) deg.

#         :param theta: Angle in degrees.
#         """
#         while (theta > 360.0) or (theta < -360.0):
#             if theta > 360.0:
#                 theta -= 360.0
#             elif theta < -360.0:
#                 theta += 360.0

#         return theta

#     # Calculate Julian century wrt to J2000
#     T = (t - 2451545.0) / 36525.0

#     # Calculate GMST in radians
#     # Algorithm from Curtis
#     a0 = 100.4606184
#     a1 = 36000.77004
#     a2 = 0.000387933
#     a3 = -2.583e-8

#     theta_GMST = a0 + a1 * T + a2 * T * T + a3 * T * T * T

#     for idx, theta in enumerate(theta_GMST):
#         theta_GMST[idx] = _constrain_theta(theta)
#     theta_GMST = np.deg2rad(theta_GMST)

#     # Calculate dates from JD

#     ut_list = np.zeros_like(t)
#     for idx, jd in enumerate(t):
#         date, err = jd2date(jd, None)
#         assert isinstance(date, Date)
#         ut_list[idx] = date.hour * 3600.0 + date.minute * 60.0 + date.second + err

#     # Calculate Greenwhich sidereal time at current epoch
#     # Earth's rotation speed taken from Horizons
#     theta_G = theta_GMST + 0.00007292115 * ut_list

#     # Rotate position vector to ECI
#     x_ECI = s.x * np.cos(theta_G) + s.y * np.sin(theta_G)
#     y_ECI = -s.x * np.sin(theta_G) + s.y * np.cos(theta_G)
#     z_ECI = s.z

#     # Calculate spherical coordinates
#     r = np.sqrt(x_ECI * x_ECI + y_ECI * y_ECI + z_ECI * z_ECI)
#     lat = np.rad2deg(np.arcsin(z_ECI / r))
#     long = np.rad2deg(np.arctan2(y_ECI, x_ECI))

#     return SphericalGeocentric(r, lat, long)

#     # return SphericalCoordinates(r, lat, long)


# def cartesian2spherical_inertial(s: CartesianState) -> SphericalGeocentric:
#     """Convert cartesian state vector to spherical, inertial coordinates

#     :param s: Cartesian state [m, m/s]
#     :return: Spherical inertial coordinates [m, deg]
#     """
#     lat = np.rad2deg(np.arcsin(s.z / s.r_mag))
#     r_xy = np.sqrt(s.x * s.x + s.y * s.y)
#     lon = np.rad2deg(np.arctan2(s.y / r_xy, s.x / r_xy))

#     return SphericalGeocentric(s.r_mag, lat, lon)



"""
Plotting submodule
==================

.. currentmodule:: nastro.plots

Generic plots
--------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    PlotSetup
    SingleAxis
    DoubleAxis
    ParasiteAxis
    Mosaic


Astrodynamics plots
--------------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    PlotCartesianState
    PlotKeplerianState
    CompareCartesianStates
    CompareKeplerianStates

"""

from .core import PlotSetup, Mosaic, SingleAxis, DoubleAxis, ParasiteAxis, Plot3D
from .astro import (
    PlotCartesianState,
    PlotKeplerianState,
    CompareCartesianStates,
    CompareKeplerianStates,
    PlotOrbit,
)

__all__ = [
    "PlotSetup",
    "Mosaic",
    "SingleAxis",
    "DoubleAxis",
    "ParasiteAxis",
    "Plot3D",
    "PlotCartesianState",
    "PlotKeplerianState",
    "CompareCartesianStates",
    "CompareKeplerianStates",
    "PlotOrbit",
]



import numpy as np
from ..types import Vector


def enrke(M: Vector, e: float | Vector, tol: float = 3.0e-15) -> Vector:
    """Calculate eccentric anomaly from mean anomaly using the ENRKE method.

    source: https://doi.org/10.1051/0004-6361/202141423

    :param M: Mean anomaly [rad]
    :param e: Eccentricity of the keplerian orbit
    :param tol: Method's accuracy (default 3.0e-15)
    :return: Eccentric anomaly [rad]
    """

    if not isinstance(e, float):
        raise NotImplementedError("Vectorized eccentricity not implemented yet")

    # TODO: Vectorize loops

    PI = 3.1415926535897932385
    TWOPI = 6.2831853071795864779

    tol2s = 2.0 * tol / (e + 2.2e-16)
    al = tol / 1.0e7
    be = tol / 0.3

    # Fit angle in range (0, 2pi) if needed

    Mr = M % TWOPI
    flip = np.ones_like(Mr, dtype=float)
    Eout = np.zeros_like(Mr, dtype=float)

    for idx, Mri in enumerate(Mr):
        if Mri > PI:
            Mr[idx] = TWOPI - Mri
        else:
            flip[idx] = -1.0

    for idx, Mri in enumerate(Mr):

        if e > 0.99 and Mri < 0.0045:
            fp = 2.7 * Mri
            fpp = 0.301
            f = 0.154
            while fpp - fp > (al + be * f):
                if (f - e * np.sin(f) - Mri) > 0.0:
                    fpp = f
                else:
                    fp = f
                f = 0.5 * (fp + fpp)
            Eout[idx] = (M[idx] + flip[idx] * (Mri - f)) % TWOPI
        else:
            Eapp = Mri + 0.999999 * Mri * (PI - Mri) / (
                2.0 * Mri + e - PI + 2.4674011002723395 / (e + 2.2e-16)
            )
            fpp = e * np.sin(Eapp)
            fppp = e * np.cos(Eapp)
            fp = 1.0 - fppp
            f = Eapp - fpp - Mri
            delta = -f / fp
            fp3 = fp * fp * fp
            ffpfpp = f * fp * fpp
            f2fppp = f * f * fppp
            delta = (
                delta
                * (fp3 - 0.5 * ffpfpp + f2fppp / 3.0)
                / (fp3 - ffpfpp + 0.5 * f2fppp)
            )

            while delta * delta > fp * tol2s:
                Eapp += delta
                fp = 1.0 - e * np.cos(Eapp)
                delta = (Mri - Eapp + e * np.sin(Eapp)) / fp

            Eapp += delta
            Eout[idx] = (M[idx] + flip[idx] * (Mri - Eapp)) % TWOPI

    return Eout



from ..types import KeplerianState, Vector, time_to_mean_anomaly, Double
from ..types.conversion import eccentric_to_true_anomaly
import numpy as np
from .enrke import enrke
from ..constants import twopi


def keplerian_orbit(s0: KeplerianState, epochs: Vector, mu: Double) -> KeplerianState:

    if not s0.scalar:
        raise ValueError("Initial state must be scalar")

    # Initial value of mean anomaly
    M = time_to_mean_anomaly(epochs - epochs[0], s0.a, mu, s0.M)
    E = enrke(M, s0.e)
    nu_wrapped = eccentric_to_true_anomaly(E, s0.e)
    nu = np.unwrap(nu_wrapped, period=twopi)
    if not np.allclose(nu[0], s0.ta, atol=1e-15, rtol=0.0):
        raise ValueError(f"True anomaly unwrapping failed: {nu[0]}, {s0.ta}")

    # Generate keplerian state
    base = np.ones_like(epochs, dtype=np.float64)
    return KeplerianState(
        s0.a * base,
        s0.e * base,
        s0.i * base,
        s0.raan * base,
        s0.aop * base,
        nu,
    )


# def keplerian_orbit(s0: KeplerianState, epochs: Vector, mu: float) -> KeplerianState:
#     """Generate an ideal keplerian trajectory

#     :param s0: Keplerian elements at initial epoch
#     :param epochs: Epoch at which to calculate the state [s]
#     :param mu: Standard gravitational parameter of central body [m^3/s^2]
#     """

#     a = np.ones_like(epochs) * s0.a
#     e = np.ones_like(epochs) * s0.e
#     i = np.ones_like(epochs) * s0.i
#     Omega = np.ones_like(epochs) * s0.Omega
#     omega = np.ones_like(epochs) * s0.omega

#     # Calculate mean anomaly as function of time
#     # M = (t - t0) * sqrt(mu / a^3)
#     M = (epochs - epochs[0]) * np.sqrt(mu / (s0.a * s0.a * s0.a))

#     # Calculate eccentric anomaly from mean anomaly
#     # M = E - e * sin(E)
#     # Newton-Raphson using M as initial guess
#     def f(E, e, M):
#         return E - e * np.sin(E) - M

#     def fprime(E, e, M):
#         return 1 - e * np.cos(E)

#     def fprime2(E, e, M):
#         return e * np.sin(E)

#     guess_E = M

#     E = newton(f, guess_E, fprime, fprime2=fprime2, args=(e, M))

#     # Calculate true anomaly from eccentric anomaly
#     nu = np.rad2deg(2.0 * np.arctan(np.sqrt((1 + e) / (1 - e)) * np.tan(E / 2)))

#     return KeplerianState(a, e, i, Omega, omega, nu)



from ..types.state import GenericState
from ..types.core import Vector, Double
from .reference_frames import Frame, ICRF
import numpy as np
from typing import Self


class CartesianPosition[T: (Double, Vector)](GenericState):
    """Three dimensional position vector in cartesian coordinates

    Components might be floating point numbers (single vector) or numpy arrays
    of them (time series of vectors).

    Parameters
    ----------
    x, y, z : Double | Vector
        Components of the position vector
    frame : Frame, optional
        Reference frame in which the position vector is defined

    Attributes
    ----------
    mag : Double | Vector
        Magnitude of the position vector
    """

    def __init__(self, x: T, y: T, z: T, *__args, frame: type[Frame] = ICRF) -> None:

        null = 0.0 * x
        super().__init__(x, y, z, null, null, null)
        self.frame = frame

        return None

    @property
    def x(self) -> T:
        return self.q1[0] if self.scalar else self.q1

    @property
    def y(self) -> T:
        return self.q2[0] if self.scalar else self.q2

    @property
    def z(self) -> T:
        return self.q3[0] if self.scalar else self.q3

    def asarray(self) -> Vector:
        return np.array([self.x, self.y, self.z])

    @property
    def mag(self) -> T:
        return np.linalg.norm(self.asarray(), axis=0)

    def __repr__(self) -> str:
        return (
            f"Position vector in {self.frame.__name__}: \n"
            "-------------------------------------------\n"
            f"x: {self.x}\n"
            f"y: {self.y}\n"
            f"z: {self.z}\n"
        )

    def convert(self, frame: type[Frame]) -> "CartesianPosition":
        """Convert position vector to a different reference frame

        Parameters
        ----------
        frame : Frame
            Reference frame in which the position vector is defined

        Returns
        -------
        CartesianPosition
            Position vector in the new reference frame
        """
        q1, q2, q3 = self.frame.transform(self.x, self.y, self.z, frame=frame)
        return CartesianPosition(q1, q2, q3, frame=frame)

    def to_spherical(self) -> "SphericalPosition":
        """Convert position vector to spherical coordinates

        Returns
        -------
        SphericalPosition
            Position vector in spherical coordinates
        """
        return SphericalPosition(
            *self.frame.to_spherical(self.x, self.y, self.z), frame=self.frame
        )


class CartesianVelocity[T: (Double, Vector)](GenericState):
    """Three dimensional velocity vector in cartesian coordinates

    Components might be floating point numbers (single vector) or numpy arrays
    of them (time series of vectors).

    Parameters
    ----------
    dx, dy, dz : Double | Vector
        Components of the velocity vector
    frame : Frame, optional
        Reference frame in which the velocity vector is defined

    Attributes
    ----------
    mag : Double | Vector
        Magnitude of the velocity vector
    """

    def __init__(self, dx: T, dy: T, dz: T, frame: type[Frame] = ICRF) -> None:

        null = 0.0 * dx
        super().__init__(null, null, null, dx, dy, dz)
        self.frame = frame

        return None

    @property
    def dx(self) -> T:
        return self.q4[0] if self.scalar else self.q4

    @property
    def dy(self) -> T:
        return self.q5[0] if self.scalar else self.q5

    @property
    def dz(self) -> T:
        return self.q6[0] if self.scalar else self.q6

    def asarray(self) -> Vector:
        return np.array([self.dx, self.dy, self.dz])

    @property
    def mag(self) -> T:
        return np.linalg.norm(self.asarray(), axis=0)

    def __repr__(self) -> str:
        return (
            f"Velocity vector in {self.frame.__name__}: \n"
            "-------------------------------------------\n"
            f"dx: {self.dx}\n"
            f"dy: {self.dy}\n"
            f"dz: {self.dz}\n"
        )


class SphericalPosition[T: (Double, Vector)](GenericState):
    """Three dimensional position vector in spherical coordinates

    Components might be floating point numbers (single vector) or numpy arrays
    of them (time series of vectors).

    Angles can be passed in radians (default) or degrees (deg=True), but they
    are always converted to radians during initialization. For each angle, the
    method SphericalPosition.angle will return the value in radians, while
    SphericalPosition.angle_deg will return it in degrees.
    """

    def __init__(
        self,
        r: T,
        theta: T,
        phi: T,
        frame: type[Frame] = ICRF,
        deg: bool = False,
        wrap: bool = False,
    ) -> None:

        null = 0.0 * r
        super().__init__(r, theta, phi, null, null, null)
        self.frame = frame

        if deg:
            self.q2 = np.deg2rad(self.q2)
            self.q3 = np.deg2rad(self.q3)
        if wrap:
            raise NotImplementedError(
                "Angle wrapping not implemented for spherical coordinates"
            )

        return None

    def to_cartesian(self) -> "CartesianPosition":
        """Convert position vector to cartesian coordinates

        Returns
        -------
        CartesianPosition
            Position vector in cartesian coordinates
        """
        return CartesianPosition(
            *self.frame.to_cartesian(self.q1, self.q2, self.q3), frame=self.frame
        )

    def __getattr__(self, name: str) -> T:

        if name in self.frame.variables:
            qi = self.__dict__[self.frame.variables[name]]
            return qi[0] if self.scalar else qi

        raise AttributeError(
            f"Attribute {name} is not defined for spherical position vector in"
            f" {self.frame.__name__}"
        )
